//-----------------------------------------------------------------------------
// File: CGameApp.cpp
//
// Desc: Game Application class, this is the central hub for all app processing
//
// Copyright (c) 1997-2002 Adam Hoult & Gary Simmons. All rights reserved.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// CGameApp Specific Includes
//-----------------------------------------------------------------------------
#include "stdafx.h"
#include "CGameApp.h"
#include "CCamera.h"

//-----------------------------------------------------------------------------
// CGameApp Member Functions
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Name : CGameApp () (Constructor)
// Desc : CGameApp Class Constructor
//-----------------------------------------------------------------------------
CGameApp::CGameApp()
{
	// Reset / Clear all required values
    m_hWnd          = NULL;
    m_pD3D          = NULL;
    m_pD3DDevice    = NULL;
    m_bLostDevice   = false;

}

//-----------------------------------------------------------------------------
// Name : ~CGameApp () (Destructor)
// Desc : CGameApp Class Destructor
//-----------------------------------------------------------------------------
CGameApp::~CGameApp()
{
	// Shut the engine down
    ShutDown();
}

//-----------------------------------------------------------------------------
// Name : InitInstance ()
// Desc : Initialises the entire Engine here.
//-----------------------------------------------------------------------------
bool CGameApp::InitInstance(HWND hwnd, CRect rc)
{
	this->m_hWnd = hwnd;
    // Initialize Direct3D (Simple)
    if (!InitDirect3D(rc)) return false;
	
    // Build Objects
    if (!BuildObjects()) { ShutDown(); return false; }

    // Set up all required game states
    SetupGameState();

    // Setup our rendering environment
    SetupRenderStates();

    // Success!
	return true;
}

//-----------------------------------------------------------------------------
// Name : InitDirect3D () (Private)
// Desc : Performs a simple, non-enumerated, initialization of Direct3D
//-----------------------------------------------------------------------------
bool CGameApp::InitDirect3D(CRect rc)
{
    HRESULT               hRet;
    D3DPRESENT_PARAMETERS PresentParams;
    D3DCAPS9              Caps;
	D3DDISPLAYMODE        CurrentMode;
	
    // First of all create our D3D Object
    m_pD3D = Direct3DCreate9( D3D_SDK_VERSION );
    if (!m_pD3D) 
    {
        MessageBox( m_hWnd, _T("No compatible Direct3D object could be created."), _T("Fatal Error!"), MB_OK | MB_ICONSTOP | MB_APPLMODAL );
        return false;
    
    } // End if failure

    // Fill out a simple set of present parameters
    ZeroMemory( &PresentParams, sizeof(D3DPRESENT_PARAMETERS) );

    // Select back buffer format etc
	m_pD3D->GetAdapterDisplayMode( D3DADAPTER_DEFAULT, &CurrentMode);
	PresentParams.BackBufferFormat = CurrentMode.Format;
        
	// Setup remaining flags
	PresentParams.AutoDepthStencilFormat = FindDepthStencilFormat( D3DADAPTER_DEFAULT, CurrentMode, D3DDEVTYPE_HAL );
    PresentParams.SwapEffect			 = D3DSWAPEFFECT_DISCARD;
    PresentParams.PresentationInterval   = D3DPRESENT_INTERVAL_IMMEDIATE;
	PresentParams.Windowed				 = true;
	PresentParams.EnableAutoDepthStencil = true;
	PresentParams.BackBufferWidth = rc.Width();
	PresentParams.BackBufferHeight = rc.Height();
	// Set Creation Flags
	unsigned long ulFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

    // Check if Hardware T&L is available
    ZeroMemory( &Caps, sizeof(D3DCAPS9) );
    m_pD3D->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &Caps );
    if ( Caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) ulFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
        
    // Attempt to create a HAL device
    if( FAILED( hRet = m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, m_hWnd, ulFlags, &PresentParams, &m_pD3DDevice ) ) ) 
    {
        MessageBox( m_hWnd, _T("Could not create a valid HAL Direct3D device object.\r\n\r\n")
                            _T("The system will now attempt to create a device utilising the 'Reference Rasterizer' (D3DDEVTYPE_REF)"),
                            _T("Fatal Error!"), MB_OK | MB_ICONINFORMATION | MB_APPLMODAL );
        
        // Find REF depth buffer format
        PresentParams.AutoDepthStencilFormat = FindDepthStencilFormat( D3DADAPTER_DEFAULT, CurrentMode, D3DDEVTYPE_REF );

        // Check if Hardware T&L is available
        ZeroMemory( &Caps, sizeof(D3DCAPS9) );
        ulFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
        m_pD3D->GetDeviceCaps( D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, &Caps );
        if ( Caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT ) ulFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;

        // Attempt to create a REF device
        if( FAILED( hRet = m_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_REF, m_hWnd, ulFlags, &PresentParams, &m_pD3DDevice ) ) ) 
        {
            MessageBox( m_hWnd, _T("Could not create a valid REF Direct3D device object.\r\n\r\nThe system will now exit."),
                                _T("Fatal Error!"), MB_OK | MB_ICONSTOP | MB_APPLMODAL );

            // Failed
            return false;
        
        } // End if Failure (REF)

    } // End if Failure (HAL)

    // Store the present parameters
    m_D3DPresentParams = PresentParams;

    // Success!!
    return true;
}

//-----------------------------------------------------------------------------
// Name : FindDepthStencilFormat ()
// Desc : This function simply determines the best depth format that is
//        available for the specified mode.
// Note : No tests for stencil active depth buffers are made.
//-----------------------------------------------------------------------------
D3DFORMAT CGameApp::FindDepthStencilFormat( ULONG AdapterOrdinal, D3DDISPLAYMODE Mode, D3DDEVTYPE DevType )
{

    // Test for 24 bith depth buffer
    if (SUCCEEDED( m_pD3D->CheckDeviceFormat(AdapterOrdinal, DevType, Mode.Format, D3DUSAGE_DEPTHSTENCIL , D3DRTYPE_SURFACE , D3DFMT_D32 )))
    {
        if (SUCCEEDED( m_pD3D->CheckDepthStencilMatch ( AdapterOrdinal, DevType, Mode.Format, Mode.Format, D3DFMT_D32 ))) return D3DFMT_D32;
    
    } // End if 32bpp Available

    // Test for 24 bit depth buffer
    if (SUCCEEDED( m_pD3D->CheckDeviceFormat(AdapterOrdinal, DevType, Mode.Format, D3DUSAGE_DEPTHSTENCIL , D3DRTYPE_SURFACE , D3DFMT_D24X8 )))
    {
        if (SUCCEEDED( m_pD3D->CheckDepthStencilMatch ( AdapterOrdinal, DevType, Mode.Format, Mode.Format, D3DFMT_D24X8 ))) return D3DFMT_D24X8;
    
    } // End if 24bpp Available

    // Test for 16 bit depth buffer
    if (SUCCEEDED( m_pD3D->CheckDeviceFormat(AdapterOrdinal, DevType, Mode.Format, D3DUSAGE_DEPTHSTENCIL , D3DRTYPE_SURFACE , D3DFMT_D16 )))
    {
        if (SUCCEEDED( m_pD3D->CheckDepthStencilMatch ( AdapterOrdinal, DevType, Mode.Format, Mode.Format, D3DFMT_D16 ))) return D3DFMT_D16;
    
    } // End if 16bpp Available

    // No depth buffer supported
    return D3DFMT_UNKNOWN;

}

//-----------------------------------------------------------------------------
// Name : SetupGameState ()
// Desc : Sets up all the initial states required by the game.
//-----------------------------------------------------------------------------
void CGameApp::SetupGameState()
{
    // Generate an identity matrix
    D3DXMatrixIdentity( &m_mtxIdentity );

    // App is active
    m_bActive = true;

    // Setup the players camera, and extract the pointer.
    // This pointer will only ever become invalid on subsequent
    // calls to CPlayer::SetCameraMode and on player destruction.
	// Setup Player details
	m_Player.SetFriction      ( 250.0f ); // Per Second
	m_Player.SetGravity       ( D3DXVECTOR3( 0, -400.0f, 0 ) );
	m_Player.SetMaxVelocityXZ ( 125.0f );
	m_Player.SetMaxVelocityY  ( 400.0f );
	m_Player.SetCamOffset     ( D3DXVECTOR3( 0.0f, 40.0f, -60.0f ) );
	m_Player.SetCamLag        ( 0.25f ); // 1/4 second camera lag

	// Switch camera mode
	m_Player.SetCameraMode    ( CCamera::MODE_THIRDPERSON );
	m_pCamera = m_Player.GetCamera();

/*  
    m_Player.SetCameraMode( CCamera::MODE_FPS );
    m_pCamera = m_Player.GetCamera();

    // Setup our player's default details
    m_Player.SetFriction( 250.0f ); // Per Second
    m_Player.SetGravity( D3DXVECTOR3( 0, -400.0f, 0 ) );
    m_Player.SetMaxVelocityXZ( 125.0f );
    m_Player.SetMaxVelocityY ( 400.0f );
    m_Player.SetCamOffset( D3DXVECTOR3( 0.0f, 10.0f, 0.0f ) );
    m_Player.SetCamLag( 0.0f );
  
	// Setup Player details
	m_Player.SetFriction      ( 250.0f ); // Per Second
	m_Player.SetGravity       ( D3DXVECTOR3( 0, -400.0f, 0 ) );
	m_Player.SetMaxVelocityXZ ( 125.0f );
	m_Player.SetMaxVelocityY  ( 400.0f );
	m_Player.SetCamOffset     ( D3DXVECTOR3( 0.0f, 40.0f, -60.0f ) );
	m_Player.SetCamLag        ( 0.25f ); // 1/4 second camera lag

	// Switch camera mode
	m_Player.SetCameraMode    ( CCamera::MODE_THIRDPERSON );
	m_pCamera = m_Player.GetCamera();
 
	// Setup player details
    m_Player.SetFriction      ( 125.0f ); // Per Second
    m_Player.SetGravity       ( D3DXVECTOR3( 0, 0, 0 ) );
    m_Player.SetMaxVelocityXZ ( 400.0f );
    m_Player.SetMaxVelocityY  ( 400.0f );
    m_Player.SetCamOffset     ( D3DXVECTOR3( 0.0f, 0.0f, 0.0f ) );
    m_Player.SetCamLag        ( 0.0f ); // No camera lag

    // Switch camera mode
    m_Player.SetCameraMode( CCamera::MODE_SPACECRAFT );
    m_pCamera = m_Player.GetCamera();
*/
	// Set up the players collision volume info
    VOLUME_INFO Volume;
    Volume.Min = D3DXVECTOR3( -3, -10, -3 );
    Volume.Max = D3DXVECTOR3(  3,  10,  3 );
    m_Player.SetVolumeInfo( Volume );

    // Setup our cameras view details
    m_pCamera->SetFOV( 80.0f );
    m_pCamera->SetViewport( m_nViewX, m_nViewY, m_nViewWidth, m_nViewHeight, 1.01f, 5000.0f );

    // Set the camera volume info (matches player volume)
    m_pCamera->SetVolumeInfo( Volume );

    // Add the update callbacks required
    m_Player.AddPlayerCallback( CTerrain::UpdatePlayer, (LPVOID)&m_Terrain );
    m_Player.AddCameraCallback( CTerrain::UpdateCamera, (LPVOID)&m_Terrain );
    
    // Lets give a small initial rotation and set initial position
    m_Player.SetPosition( D3DXVECTOR3( 430.0f, 400.0f, 330.0f ) );
    m_Player.Rotate( 25, 45, 0 );
}

//-----------------------------------------------------------------------------
// Name : SetupRenderStates ()
// Desc : Sets up all the initial states required by the renderer.
//-----------------------------------------------------------------------------
void CGameApp::SetupRenderStates()
{
    // Setup our D3D Device initial states
    m_pD3DDevice->SetRenderState( D3DRS_ZENABLE, D3DZB_TRUE );
    m_pD3DDevice->SetRenderState( D3DRS_DITHERENABLE,  TRUE );
    m_pD3DDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
    m_pD3DDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_CCW );
    m_pD3DDevice->SetRenderState( D3DRS_LIGHTING, FALSE );

    // Setup option dependant states
    m_pD3DDevice->SetRenderState( D3DRS_FILLMODE, m_FillMode );                

    // Setup our vertex FVF code
    m_pD3DDevice->SetFVF( D3DFVF_XYZ | D3DFVF_DIFFUSE );

        // Store new viewport sizes
    RECT rc;
    ::GetClientRect( m_hWnd, &rc );
    m_nViewWidth  = rc.right - rc.left;
    m_nViewHeight = rc.bottom - rc.top;
    
    // Update our device with our camera details (Required on reset)
    if ( !m_pCamera ) return;
    m_pCamera->SetViewport( m_nViewX, m_nViewY, m_nViewWidth, m_nViewHeight, 1.01f, 5000.0f );
    m_pCamera->UpdateRenderView( m_pD3DDevice );
    m_pCamera->UpdateRenderProj( m_pD3DDevice );
}

//-----------------------------------------------------------------------------
// Name : ShutDown ()
// Desc : Shuts down the game engine, and frees up all resources.
//-----------------------------------------------------------------------------
bool CGameApp::ShutDown()
{
	 // Release any previously built objects
    ReleaseObjects ( );

    // Destroy Direct3D Objects
    if ( m_pD3DDevice ) m_pD3DDevice->Release();
    if ( m_pD3D       ) m_pD3D->Release();
    m_pD3D          = NULL;
    m_pD3DDevice    = NULL;
    
    // Destroy menu, it may not be attached
    if ( m_hMenu ) DestroyMenu( m_hMenu );
    m_hMenu         = NULL;

    // Destroy the render window
    SetMenu( m_hWnd, NULL );
    if ( m_hWnd ) DestroyWindow( m_hWnd );
    m_hWnd          = NULL;
    
    // Shutdown Success
    return true;
}

//-----------------------------------------------------------------------------
// Name : BuildObjects ()
// Desc : Build our demonstration cube mesh, and the objects that instance it
//-----------------------------------------------------------------------------
bool CGameApp::BuildObjects()
{
    // Release previously built objects
    ReleaseObjects();
	bool HardwareTnL = true;

    // Build our terrain data
    m_Terrain.SetD3DDevice( m_pD3DDevice, HardwareTnL );
    if ( !m_Terrain.LoadHeightMap( _T("Data\\HeightMap.raw"), 257, 257 )) return false;

    // Build a 'player' mesh (this is just a cube currently)
    CVertex * pVertex = NULL;
    srand( timeGetTime() );

    // Add the 8 cube vertices to this mesh
    if ( m_PlayerMesh.AddVertex( 8 ) < 0 ) return false;

    // Add all 4 vertices
    pVertex = &m_PlayerMesh.m_pVertex[0];
    
    // Add bottom 4 vertices
    *pVertex++ = CVertex( -3, 0, -3, RANDOM_COLOR );
    *pVertex++ = CVertex( -3, 0,  3, RANDOM_COLOR );
    *pVertex++ = CVertex(  3, 0,  3, RANDOM_COLOR );
    *pVertex++ = CVertex(  3, 0, -3, RANDOM_COLOR );
    
    // Add top 4 vertices
    *pVertex++ = CVertex( -3, 20, -3, RANDOM_COLOR );
    *pVertex++ = CVertex( -3, 20,  3, RANDOM_COLOR );
    *pVertex++ = CVertex(  3, 20,  3, RANDOM_COLOR );
    *pVertex++ = CVertex(  3, 20, -3, RANDOM_COLOR );

    // Add the indices as a strip (with one degenerate) ;)
    if ( m_PlayerMesh.AddIndex( 16 ) < 0 ) return false;
    m_PlayerMesh.m_pIndex[ 0] = 5;
    m_PlayerMesh.m_pIndex[ 1] = 6;
    m_PlayerMesh.m_pIndex[ 2] = 4;
    m_PlayerMesh.m_pIndex[ 3] = 7; //               5-6
    m_PlayerMesh.m_pIndex[ 4] = 0; //               |/|
    m_PlayerMesh.m_pIndex[ 5] = 3; //               4-7
    m_PlayerMesh.m_pIndex[ 6] = 1; //               |/|
    m_PlayerMesh.m_pIndex[ 7] = 2; //               0-3-7
    m_PlayerMesh.m_pIndex[ 8] = 3; // Degen Index   |/|/|
    m_PlayerMesh.m_pIndex[ 9] = 7; //               1-2-6
    m_PlayerMesh.m_pIndex[10] = 2; //                 |/|
    m_PlayerMesh.m_pIndex[11] = 6; //                 1-5
    m_PlayerMesh.m_pIndex[12] = 1; //                 |/|
    m_PlayerMesh.m_pIndex[13] = 5; //                 0-4
    m_PlayerMesh.m_pIndex[14] = 0;
    m_PlayerMesh.m_pIndex[15] = 4;

    // Build the mesh's vertex and index buffers
    if (FAILED(m_PlayerMesh.BuildBuffers( m_pD3DDevice, HardwareTnL, true ))) return false;

    // Our object references this mesh
    m_Object.m_pMesh = &m_PlayerMesh;

    // Link this object to our player
    m_Player.Set3rdPersonObject( &m_Object );

    // Success!
    return true;
}

//-----------------------------------------------------------------------------
// Name : FrameAdvance () (Private)
// Desc : Called to signal that we are now rendering the next frame.
//-----------------------------------------------------------------------------
void CGameApp::FrameAdvance()
{
    static TCHAR FrameRate[ 50 ];
    static TCHAR TitleBuffer[ 255 ];

    // Advance the timer
    m_Timer.Tick( );

    // Skip if app is inactive
    if ( !m_bActive ) return;
    
    // Get / Display the framerate
    if ( m_LastFrameRate != m_Timer.GetFrameRate() )
    {
		CWnd *pWnd = AfxGetMainWnd();
		HWND hwnd = pWnd->m_hWnd;

        m_LastFrameRate = m_Timer.GetFrameRate( FrameRate );
        _stprintf( TitleBuffer, _T("Simple Terrain : %s"), FrameRate );
        SetWindowText( hwnd, TitleBuffer );

    } // End if Frame Rate Altered

 /*   // Recover lost device if required
    if ( m_bLostDevice )
    {
        // Can we reset the device yet ?
        HRESULT hRet = m_pD3DDevice->TestCooperativeLevel();
        if ( hRet == D3DERR_DEVICENOTRESET )
        {
            // Restore the device
            CMyD3DInit Initialize;
            Initialize.ResetDisplay( m_pD3DDevice, m_D3DSettings, m_hWnd );
            SetupRenderStates( );
            m_bLostDevice = false;

        } // End if can reset
        else
        {
            return;

        } // End if cannot reset

    } // End if Device Lost
*/
    // Poll & Process input devices
    ProcessInput();

    // Animate the game objects
    AnimateObjects();

    // Clear the frame & depth buffer ready for drawing
    m_pD3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x000000, 1.0f, 0 );
    
    // Begin Scene Rendering
    m_pD3DDevice->BeginScene();
    
    // Reset our world matrix
    m_pD3DDevice->SetTransform( D3DTS_WORLD, &m_mtxIdentity );

    // Render our terrain objects
    m_Terrain.Render( m_pCamera );

    // Request our player render itselfs
    m_Player.Render( m_pD3DDevice );

    // End Scene Rendering
    m_pD3DDevice->EndScene();
    
    // Present the buffer
    if ( FAILED(m_pD3DDevice->Present( NULL, NULL, NULL, NULL )) ) m_bLostDevice = true;

}

//-----------------------------------------------------------------------------
// Name : ProcessInput () (Private)
// Desc : Simply polls the input devices and performs basic input operations
//-----------------------------------------------------------------------------
void CGameApp::ProcessInput( )
{
   static UCHAR pKeyBuffer[ 256 ];
    ULONG        Direction = 0;
    POINT        CursorPos;
    float        X = 0.0f, Y = 0.0f;

    // Retrieve keyboard state
    if ( !GetKeyboardState( pKeyBuffer ) ) return;

    // Check the relevant keys
    if ( pKeyBuffer[ VK_UP    ] & 0xF0 ) Direction |= CPlayer::DIR_FORWARD;
    if ( pKeyBuffer[ VK_DOWN  ] & 0xF0 ) Direction |= CPlayer::DIR_BACKWARD;
    if ( pKeyBuffer[ VK_LEFT  ] & 0xF0 ) Direction |= CPlayer::DIR_LEFT;
    if ( pKeyBuffer[ VK_RIGHT ] & 0xF0 ) Direction |= CPlayer::DIR_RIGHT;
    if ( pKeyBuffer[ VK_PRIOR ] & 0xF0 ) Direction |= CPlayer::DIR_UP;
    if ( pKeyBuffer[ VK_NEXT  ] & 0xF0 ) Direction |= CPlayer::DIR_DOWN;
    
    // Now process the mouse (if the button is pressed)
    if ( GetCapture() == m_hWnd )
    {
        // Hide the mouse pointer
        SetCursor( NULL );

        // Retrieve the cursor position
        GetCursorPos( &CursorPos );

        // Calculate mouse rotational values
        X = (float)(CursorPos.x - m_OldCursorPos.x) / 3.0f;
        Y = (float)(CursorPos.y - m_OldCursorPos.y) / 3.0f;

        // Reset our cursor position so we can keep going forever :)
        SetCursorPos( m_OldCursorPos.x, m_OldCursorPos.y );

    } // End if Captured

    // Update if we have moved
    if ( Direction > 0 || X != 0.0f || Y != 0.0f )
    {
        // Rotate our camera
        if ( X || Y ) 
        {
            // Are they holding the right mouse button ?
            if ( pKeyBuffer[ VK_RBUTTON ] & 0xF0 )
                m_Player.Rotate( Y, 0.0f, -X );
            else
                m_Player.Rotate( Y, X, 0.0f );
        
        } // End if any rotation

        // Any Movement ?
        if ( Direction ) 
        {
            // Move our player (Force applied must be greater than total friction)
            m_Player.Move( Direction, 500.0f * m_Timer.GetTimeElapsed(), true );

        } // End if any movement

    } // End if camera moved

    // Update our camera (updates velocity etc)
    m_Player.Update( m_Timer.GetTimeElapsed() );

    // Update the device matrix
    m_pCamera->UpdateRenderView( m_pD3DDevice );

}

//-----------------------------------------------------------------------------
// Name : AnimateObjects () (Private)
// Desc : Animates the objects we currently have loaded.
//-----------------------------------------------------------------------------
void CGameApp::AnimateObjects()
{
    // Note : Expanded for the purposes of this example only.
   
}

//-----------------------------------------------------------------------------
// Name : ReleaseObjects ()
// Desc : Releases our objects and their associated memory so that we can
//        rebuild them, if required, during our applications life-time.
//-----------------------------------------------------------------------------
void CGameApp::ReleaseObjects( )
{
    // De-reference device objects stream (just in case any remain set)
    if ( m_pD3DDevice ) 
    {
        m_pD3DDevice->SetStreamSource( 0, NULL, 0, 0 );
        m_pD3DDevice->SetIndices( NULL );

    } // End if Device Available

    // Release any required objects
    m_Terrain.Release();
}

//-----------------------------------------------------------------------------
// MFC  ϱ  Լ
//-----------------------------------------------------------------------------
void CGameApp::Mouse_Lift_Down()
{
	// Capture the mouse
	SetCapture( m_hWnd );
	GetCursorPos( &m_OldCursorPos );

}

void CGameApp::Mouse_Lift_Up()
{
	ReleaseCapture( );
}

void CGameApp::SetSize(bool isMin, CRect rc)
{
	if (isMin)
	{
		// App is inactive
		m_bActive = false;

	} // App has been minimized
	else
	{
		// App is active
		m_bActive = true;

		// Store new viewport sizes
		m_nViewWidth  = rc.Width();
		m_nViewHeight = rc.Height();

		if (m_pD3DDevice) 
		{
			// Store new sizes
			m_D3DPresentParams.BackBufferWidth = m_nViewWidth;
			m_D3DPresentParams.BackBufferHeight = m_nViewHeight;

			// Reset the device
			m_pD3DDevice->Reset( &m_D3DPresentParams );
			SetupRenderStates( );

		} // End if

	} // End if !Minimized

}
